// =================================================================================================
// Copyright 2002-2005 Adobe Systems Incorporated
// All Rights Reserved.
//
// NOTICE:  Adobe permits you to use, modify, and distribute this file in accordance with the terms
// of the Adobe license agreement accompanying it.
//
// Adobe patent application tracking #P435, entitled 'Unique markers to simplify embedding data of
// one format in a file with a different format', inventors: Sean Parent, Greg Gilley.
// =================================================================================================

#include "XMP_Environment.h"	// ! This must be the first include!
#include "XMP_Impl.hpp"

#include "XMPMeta.hpp"
#include "XMPUtils.hpp"

#include "UnicodeInlines.incl_cpp"
#include "UnicodeConversions.hpp"
#include "ExpatAdapter.hpp"

#if XMP_DebugBuild
	#include <iostream>
#endif

using namespace std;

#if XMP_WinBuild
	#pragma warning ( disable : 4533 )	// initialization of '...' is skipped by 'goto ...'
	#pragma warning ( disable : 4702 )	// unreachable code
	#pragma warning ( disable : 4800 )	// forcing value to bool 'true' or 'false' (performance warning)
	#pragma warning ( disable : 4996 )	// '...' was declared deprecated
#endif


// *** Use the XMP_PropIsXyz (Schema, Simple, Struct, Array, ...) macros
// *** Add debug codegen checks, e.g. that typical masking operations really work
// *** Change all uses of strcmp and strncmp to XMP_LitMatch and XMP_LitNMatch


// =================================================================================================
// Local Types and Constants
// =========================


// =================================================================================================
// Static Variables
// ================


// =================================================================================================
// Local Utilities
// ===============


#define IsHexDigit(ch)		( (('0' <= (ch)) && ((ch) <= '9')) || (('A' <= (ch)) && ((ch) <= 'F')) )
#define HexDigitValue(ch)	( (((ch) - '0') < 10) ? ((ch) - '0') : ((ch) - 'A' + 10) )


// -------------------------------------------------------------------------------------------------
// PickBestRoot
// ------------
//
// Pick the first pxmp:XMP_Packet or x:xmpmeta among multiple root candidates. If there aren't any,
// pick the first bare rdf:RDF if that is allowed. The returned root is the rdf:RDF child if an
// x:xmpmeta element was chosen.  The search is breadth first, so a higher level candiate is chosen
// over a lower level one that was textually earlier in the serialized XML.

static const XML_Node * PickBestRoot ( const XML_Node & xmlParent, XMP_OptionBits options )
{
	// Look among this parent's content for pxmp:XMP_Packet or x:xmpmeta. The recursion for
	// x:xmpmeta is broader than the strictly defined choice, but gives us smaller code.
	for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
		const XML_Node * childNode = xmlParent.content[childNum];
		if ( childNode->kind != kElemNode ) continue;
		if ( childNode->name == "pxmp:XMP_Packet" ) return childNode;
		if ( (childNode->name == "x:xmpmeta") || (childNode->name == "x:xapmeta") ) return PickBestRoot ( *childNode, 0 );
	}
	
	// Look among this parent's content for a bare rdf:RDF if that is allowed.
	if ( ! (options & kXMP_RequireXMPMeta) ) {
		for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
			const XML_Node * childNode = xmlParent.content[childNum];
			if ( childNode->kind != kElemNode ) continue;
			if ( childNode->name == "rdf:RDF" ) return childNode;
		}
	}
	
	// Recurse into the content.
	for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
		const XML_Node * foundRoot = PickBestRoot ( *xmlParent.content[childNum], options );
		if ( foundRoot != 0 ) return foundRoot;
	}
	
	return 0;

}	// PickBestRoot

// -------------------------------------------------------------------------------------------------
// FindRootNode
// ------------
//
// Find the XML node that is the root of the XMP data tree. Generally this will be an outer node,
// but it could be anywhere if a general XML document is parsed (e.g. SVG). The XML parser counted
// all rdf:RDF and pxmp:XMP_Packet nodes, and kept a pointer to the last one. If there is more than
// one possible root use PickBestRoot to choose among them.
//
// If there is a root node, try to extract the version of the previous XMP toolkit.

static const XML_Node * FindRootNode ( XMPMeta * thiz, const XMLParserAdapter & xmlParser, XMP_OptionBits options )
{
	const XML_Node * rootNode = xmlParser.rootNode;
	
	if ( xmlParser.rootCount > 1 ) rootNode = PickBestRoot ( xmlParser.tree, options );
	if ( rootNode == 0 ) return 0;
	
	// We have a root node. Try to extract previous toolkit version number.
	
	XMP_StringPtr verStr = "";
	
	if ( rootNode->name == "pxmp:XMP_Packet" ) {

		for ( size_t attrNum = 0, attrLim = rootNode->attrs.size(); attrNum < attrLim; ++attrNum ) {
			const XML_Node * currAttr =rootNode->attrs[attrNum];
			if ( currAttr->name == "pxmp:xmptk" ) {
				verStr = currAttr->value.c_str();
				break;
			}
		}

	} else {
	
		XMP_Assert ( rootNode->name == "rdf:RDF" );
	
		if ( (options & kXMP_RequireXMPMeta) &&
		     ((rootNode->parent == 0) ||
		      ((rootNode->parent->name != "x:xmpmeta") && (rootNode->parent->name != "x:xapmeta"))) ) return 0;

		for ( size_t attrNum = 0, attrLim = rootNode->parent->attrs.size(); attrNum < attrLim; ++attrNum ) {
			const XML_Node * currAttr =rootNode->parent->attrs[attrNum];
			if ( (currAttr->name == "x:xmptk") || (currAttr->name == "x:xaptk") ) {
				verStr = currAttr->value.c_str();
				break;
			}
		}
		
	}
	
	// Decode the version number into MMmmuubbb digits. If any part is too big, peg it at 99 or 999.
	
	unsigned long part;
	while ( (*verStr != 0) && ((*verStr < '0') || (*verStr > '9')) ) ++verStr;
	
	part = 0;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer = part * 100*100*1000;
	
	part = 0;
	if ( *verStr == '.' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer += part * 100*1000;
	
	part = 0;
	if ( *verStr == '.' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer += part * 1000;
	
	part = 0;
	if ( *verStr == '-' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 999 ) part = 999;
	thiz->prevTkVer += part;
	
	return rootNode;
	
}	// FindRootNode

// -------------------------------------------------------------------------------------------------
// NormalizeDCArrays
// -----------------
//
// Undo the denormalization performed by the XMP used in Acrobat 5. If a Dublin Core array had only
// one item, it was serialized as a simple property. The xml:lang attribute was dropped from an
// alt-text item if the language was x-default.

// *** This depends on the dc: namespace prefix.

static void
NormalizeDCArrays ( XMP_Node * xmpTree )
{
	XMP_Node * dcSchema = FindSchemaNode ( xmpTree, kXMP_NS_DC, kXMP_ExistingOnly );
	if ( dcSchema == 0 ) return;
	
	for ( size_t propNum = 0, propLimit = dcSchema->children.size(); propNum < propLimit; ++propNum ) {
		XMP_Node *     currProp  = dcSchema->children[propNum];
		XMP_OptionBits arrayForm = 0;
		
		if ( ! XMP_PropIsSimple ( currProp->options ) ) continue;	// Nothing to do if not simple.
		
		if ( (currProp->name == "dc:creator" )     ||	// See if it is supposed to be an array.
		     (currProp->name == "dc:date" ) ) {			// *** Think about an array of char* and a loop.
			arrayForm = kXMP_PropArrayIsOrdered;
		} else if (
		     (currProp->name == "dc:description" ) ||
		     (currProp->name == "dc:rights" )      ||
		     (currProp->name == "dc:title" ) ) {
			arrayForm = kXMP_PropArrayIsAltText;
		} else if (
		     (currProp->name == "dc:contributor" ) ||
		     (currProp->name == "dc:language" )    ||
		     (currProp->name == "dc:publisher" )   ||
		     (currProp->name == "dc:relation" )    ||
		     (currProp->name == "dc:subject" )     ||
		     (currProp->name == "dc:type" ) ) {
			arrayForm = kXMP_PropValueIsArray;
		}
		if ( arrayForm == 0 ) continue;	// Nothing to do if it isn't supposed to be an array.
		
		arrayForm = VerifySetOptions ( arrayForm, 0 );	// Set the implicit array bits.
		XMP_Node * newArray = new XMP_Node ( dcSchema, currProp->name.c_str(), arrayForm );
		dcSchema->children[propNum] = newArray;
		newArray->children.push_back ( currProp );
		currProp->parent = newArray;
		currProp->name = kXMP_ArrayItemName;
		
		if ( XMP_ArrayIsAltText ( arrayForm ) && (! (currProp->options & kXMP_PropHasLang)) ) {
			XMP_Node * newLang = new XMP_Node ( currProp, "xml:lang", "x-default", kXMP_PropIsQualifier );
			currProp->options |= (kXMP_PropHasQualifiers | kXMP_PropHasLang);
			if ( currProp->qualifiers.empty() ) {	// *** Need a util?
				currProp->qualifiers.push_back ( newLang );
			} else {
				currProp->qualifiers.insert ( currProp->qualifiers.begin(), newLang );
			}
		}

	}
	
}	// NormalizeDCArrays


// -------------------------------------------------------------------------------------------------
// CompareAliasedSubtrees
// ----------------------

// *** Change to do some alias-specific setup, then use CompareSubtrees. One special case for
// *** aliases is a simple to x-default alias, the options and qualifiers obviously differ.

static void
CompareAliasedSubtrees ( XMP_Node * aliasNode, XMP_Node * baseNode, bool outerCall = true )
{
	// ! The outermost call is special. The names almost certainly differ. The qualifiers (and
	// ! hence options) will differ for an alias to the x-default item of a langAlt array.
	if ( (aliasNode->value != baseNode->value) ||
	     (aliasNode->children.size() != baseNode->children.size()) ) {
		XMP_Throw ( "Mismatch between alias and base nodes", kXMPErr_BadXMP );
	}
	if ( ! outerCall ) {
		if ( (aliasNode->name != baseNode->name) ||
		     (aliasNode->options != baseNode->options) ||
		     (aliasNode->qualifiers.size() != baseNode->qualifiers.size()) ) {
			XMP_Throw ( "Mismatch between alias and base nodes", kXMPErr_BadXMP );
		}
	}
	
	for ( size_t childNum = 0, childLim = aliasNode->children.size(); childNum < childLim; ++childNum ) {
		XMP_Node * aliasChild = aliasNode->children[childNum];
		XMP_Node * baseChild  = baseNode->children[childNum];
		CompareAliasedSubtrees ( aliasChild, baseChild, false );
	}
	
	for ( size_t qualNum = 0, qualLim = aliasNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
		XMP_Node * aliasQual = aliasNode->qualifiers[qualNum];
		XMP_Node * baseQual  = baseNode->qualifiers[qualNum];
		CompareAliasedSubtrees ( aliasQual, baseQual, false );
	}
	
}	// CompareAliasedSubtrees


// -------------------------------------------------------------------------------------------------
// TransplantArrayItemAlias
// ------------------------

static void
TransplantArrayItemAlias ( XMP_Node * oldParent, size_t oldNum, XMP_Node * newParent )
{
	XMP_Node * childNode = oldParent->children[oldNum];

	if ( newParent->options & kXMP_PropArrayIsAltText ) {
		if ( childNode->options & kXMP_PropHasLang ) {
			XMP_Throw ( "Alias to x-default already has a language qualifier", kXMPErr_BadXMP );	// *** Allow x-default.
		}
		childNode->options |= (kXMP_PropHasQualifiers | kXMP_PropHasLang);
		XMP_Node * langQual = new XMP_Node ( childNode, "xml:lang", "x-default", kXMP_PropIsQualifier );	// *** AddLangQual util?
		if ( childNode->qualifiers.empty() ) {
			childNode->qualifiers.push_back ( langQual );
		} else {
			childNode->qualifiers.insert ( childNode->qualifiers.begin(), langQual );
		}
	}

	oldParent->children.erase ( oldParent->children.begin() + oldNum );
	childNode->name = kXMP_ArrayItemName;
	childNode->parent = newParent;
	if ( newParent->children.empty() ) {
		newParent->children.push_back ( childNode );
	} else {
		newParent->children.insert ( newParent->children.begin(), childNode );
	}

}	// TransplantArrayItemAlias


// -------------------------------------------------------------------------------------------------
// TransplantNamedAlias
// --------------------

static void
TransplantNamedAlias ( XMP_Node * oldParent, size_t oldNum, XMP_Node * newParent, XMP_VarString & newName )
{
	XMP_Node * childNode = oldParent->children[oldNum];

	oldParent->children.erase ( oldParent->children.begin() + oldNum );
	childNode->name = newName;
	childNode->parent = newParent;
	newParent->children.push_back ( childNode );

}	// TransplantNamedAlias


// -------------------------------------------------------------------------------------------------
// MoveExplicitAliases
// -------------------

static void
MoveExplicitAliases ( XMP_Node * tree, XMP_OptionBits parseOptions )
{
	tree->options ^= kXMP_PropHasAliases;
	const bool strictAliasing = ((parseOptions & kXMP_StrictAliasing) != 0);
	
	// Visit all of the top level nodes looking for aliases. If there is no base, transplant the
	// alias subtree. If there is a base and strict aliasing is on, make sure the alias and base 
	// subtrees match.
	
	// ! Use "while" loops not "for" loops since both the schema and property loops can remove the
	// ! current item from the vector being traversed. And don't increment the counter for a delete.
	
	size_t schemaNum = 0;
	while ( schemaNum < tree->children.size() ) {
		XMP_Node * currSchema = tree->children[schemaNum];
		
		size_t propNum = 0;
		while ( propNum < currSchema->children.size() ) {
			XMP_Node * currProp = currSchema->children[propNum];
			if ( ! (currProp->options & kXMP_PropIsAlias) ) {
				++propNum;
				continue;
			}
			currProp->options ^= kXMP_PropIsAlias;

			// Find the base path, look for the base schema and root node.

			XMP_AliasMapPos aliasPos = sRegisteredAliasMap->find ( currProp->name );
			XMP_Assert ( aliasPos != sRegisteredAliasMap->end() );
			XMP_ExpandedXPath & basePath = aliasPos->second;
			XMP_OptionBits arrayOptions = (basePath[kRootPropStep].options & kXMP_PropArrayFormMask);

			XMP_Node * baseSchema = FindSchemaNode ( tree, basePath[kSchemaStep].step.c_str(), kXMP_CreateNodes );
			if ( baseSchema->options & kXMP_NewImplicitNode ) baseSchema->options ^= kXMP_NewImplicitNode;
			XMP_Node * baseNode = FindChildNode ( baseSchema, basePath[kRootPropStep].step.c_str(), kXMP_ExistingOnly );

			if ( baseNode == 0 ) {
			
				if ( basePath.size() == 2 ) {
					// A top-to-top alias, transplant the property.
					TransplantNamedAlias ( currSchema, propNum, baseSchema, basePath[kRootPropStep].step );
				} else {
					// An alias to an array item, create the array and transplant the property.
					baseNode = new XMP_Node ( baseSchema, basePath[kRootPropStep].step.c_str(), arrayOptions );
					baseSchema->children.push_back ( baseNode );
					TransplantArrayItemAlias ( currSchema, propNum, baseNode );
				}
			
			} else if ( basePath.size() == 2 ) {
			
				// The base node does exist and this is a top-to-top alias. Check for conflicts if
				// strict aliasing is on. Remove and delete the alias subtree.
				if ( strictAliasing ) CompareAliasedSubtrees ( currProp, baseNode );
				currSchema->children.erase ( currSchema->children.begin() + propNum );
				delete currProp;
			
			} else {
			
				// This is an alias to an array item and the array exists. Look for the aliased item.
				// Then transplant or check & delete as appropriate.
				
				XMP_Node * itemNode = 0;
				if ( arrayOptions & kXMP_PropArrayIsAltText ) {
					XMP_Index xdIndex = LookupLangItem ( baseNode, *xdefaultName );
					if ( xdIndex != -1 ) itemNode = baseNode->children[xdIndex];
				} else if ( ! baseNode->children.empty() ) {
					itemNode = baseNode->children[0];
				}
				
				if ( itemNode == 0 ) {
					TransplantArrayItemAlias ( currSchema, propNum, baseNode );
				} else {
					if ( strictAliasing ) CompareAliasedSubtrees ( currProp, itemNode );
					currSchema->children.erase ( currSchema->children.begin() + propNum );
					delete currProp;
				}

			}

		}	// Property loop
		
		// Increment the counter or remove an empty schema node.
		if ( currSchema->children.size() > 0 ) {
			++schemaNum;
		} else {
			delete tree->children[schemaNum];	// ! Delete the schema node itself.
			tree->children.erase ( tree->children.begin() + schemaNum );
		}
		
	}	// Schema loop
	
}	// MoveExplicitAliases


// -------------------------------------------------------------------------------------------------
// FixASCIIControls
// ----------------
//
// Undo the "doubled escaping" of ASCII control characters. Look for "&#xnn;", where "nn" is one or
// two hex digits. Where found, replace with the single byte whose value is nn.
//
// It is a bit sad to walk the entire tree to do this, but doing it in the parsing code it risky.
// Doing it in the XML parsing might improperly process text that is not actually doubled escapes.
// Doing it in the RDF recognizer is complex, hard to make sure a value is checked exactly once.

static void FixASCIIControls ( XMP_VarString & strValue )
{
	if ( strValue.size() < 5 ) return;
	XMP_StringPtr cStr = strValue.c_str();	// ! Need the terminating null to ensure cStr[5] is valid.

	for ( size_t i = 0; i+5 <= strValue.size(); ++i ) {	// ! Need dynamic limit, strValue.size() can change.

		if ( (cStr[i] == '&') && (cStr[i+1] == '#') && (cStr[i+2] == 'x') && IsHexDigit(cStr[i+3]) ) {

			int excessLen = 4;
			unsigned char chValue = HexDigitValue ( cStr[i+3] );

			if ( IsHexDigit(cStr[i+4]) ) {
				++excessLen;
				chValue = (chValue << 4) + HexDigitValue(cStr[i+4]);
			}
			if ( cStr[i+excessLen] != ';' ) continue;	// Not a complete escape sequence.

			strValue[i] = chValue;
			strValue.erase ( i+1, excessLen );
			cStr = strValue.c_str();	// Probably doesn't change, but let's be safe.

		}

	}
	
}	// FixASCIIControls


// -------------------------------------------------------------------------------------------------
// FixGPSTimeStamp
// ---------------

static void
FixGPSTimeStamp ( XMP_Node * exifSchema, XMP_Node * gpsDateTime )
{
	XMP_DateTime binGPSStamp;
	try {
		XMPUtils::ConvertToDate ( gpsDateTime->value.c_str(), &binGPSStamp );
	} catch ( ... ) {
		return;	// Don't let a bad date stop other things.
	}
	if ( (binGPSStamp.year != 0) || (binGPSStamp.month != 0) || (binGPSStamp.day != 0) ) return;
	
	XMP_Node * otherDate = FindChildNode ( exifSchema, "exif:DateTimeOriginal", kXMP_ExistingOnly );
	if ( otherDate == 0 ) otherDate = FindChildNode ( exifSchema, "exif:DateTimeDigitized", kXMP_ExistingOnly );
	if ( otherDate == 0 ) return;

	XMP_DateTime binOtherDate;
	try {
		XMPUtils::ConvertToDate ( otherDate->value.c_str(), &binOtherDate );
	} catch ( ... ) {
		return;	// Don't let a bad date stop other things.
	}
	
	binGPSStamp.year  = binOtherDate.year;
	binGPSStamp.month = binOtherDate.month;
	binGPSStamp.day   = binOtherDate.day;

	XMP_StringPtr goodStr;
	XMP_StringLen goodLen;
	XMPUtils::ConvertFromDate ( binGPSStamp, &goodStr, &goodLen );
	
	gpsDateTime->value.assign ( goodStr, goodLen );

}	// FixGPSTimeStamp


// -------------------------------------------------------------------------------------------------
// TouchUpPropNode
// ---------------

static void
TouchUpPropNode ( XMP_Node * currNode )
{
	// This is a property or qualifier. Recurse on the children of structs and arrays, and the
	// qualifiers for all nodes. Fix up the ASCII controls in the values of simple leaf nodes.
	
	// *** We should only do this for RDF serialized by 3.1-053 or later. This awaits pulling
	// *** version info from the x:xmpmeta element. Doing it all the time will change user data
	// *** of "&#x9;" into a tab because old toolkits didn't double escape when serializing.

	if ( XMP_PropIsSimple ( currNode->options ) ) {
		FixASCIIControls ( currNode->value );
	} else {
		for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {
			XMP_Node * currChild = currNode->children[childNum];
			TouchUpPropNode ( currChild );
		}
	}
	
	for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
		XMP_Node * currQual = currNode->qualifiers[qualNum];
		TouchUpPropNode ( currQual );
	}

}	// TouchUpPropNode


// -------------------------------------------------------------------------------------------------
// TouchUpOneSchema
// ----------------

static void
TouchUpOneSchema ( XMP_Node * currSchema )
{
	// Visit all of the properties to do general fixes.
	for ( size_t propNum = 0, propLim = currSchema->children.size(); propNum < propLim; ++propNum ) {
		XMP_Node * currProp = currSchema->children[propNum];
		TouchUpPropNode ( currProp );
	}
	
	if ( currSchema->name == kXMP_NS_DC ) {
		// Do a special case fix for dc:subject, make sure it isn't an ordered array.
		XMP_Node * dcSubject = FindChildNode ( currSchema, "dc:subject", kXMP_ExistingOnly );
		if ( dcSubject != 0 ) {
			XMP_OptionBits keepMask = ~(kXMP_PropArrayIsOrdered | kXMP_PropArrayIsAlternate | kXMP_PropArrayIsAltText);
			dcSubject->options &= keepMask;	// Make sure any ordered array bits are clear.
		}
	} else if ( currSchema->name == kXMP_NS_EXIF ) {
		// Do a special case fix for exif:GPSTimeStamp.
		XMP_Node * gpsDateTime = FindChildNode ( currSchema, "exif:GPSTimeStamp", kXMP_ExistingOnly );
		if ( gpsDateTime != 0 ) FixGPSTimeStamp ( currSchema, gpsDateTime );
	}

}	// TouchUpOneSchema


// -------------------------------------------------------------------------------------------------
// TouchUpDataModel
// ----------------

static void
TouchUpDataModel ( XMP_Node * tree )
{
	// Do the special case fixes within each schema.
	for ( size_t schemaNum = 0, schemaLim = tree->children.size(); schemaNum < schemaLim; ++schemaNum ) {
		XMP_Node * currSchema = tree->children[schemaNum];
		TouchUpOneSchema ( currSchema );
	}
	
	// Tweak old XMP: Move an instance ID from rdf:about to the xmpMM:InstanceID property. An old
	// instance ID usually looks like "uuid:bac965c4-9d87-11d9-9a30-000d936b79c4", plus InDesign
	// 3.0 wrote them like "bac965c4-9d87-11d9-9a30-000d936b79c4". If the name looks like a UUID
	// simply move it to xmpMM:InstanceID, don't worry about any existing xmpMM:InstanceID. Both
	// will only be present when a newer file with the xmpMM:InstanceID property is updated by an
	// old app that uses rdf:about.
	
	if ( ! tree->name.empty() ) {

		bool nameIsUUID = false;
		XMP_StringPtr nameStr = tree->name.c_str();

		if ( XMP_LitNMatch ( nameStr, "uuid:", 5 ) ) {

			nameIsUUID = true;

		} else if ( tree->name.size() == 36 ) {

			nameIsUUID = true;	// ! Assume true, we'll set it to false below if not.
			for ( int i = 0;  i < 36; ++i ) {
				char ch = nameStr[i];
				if ( ch == '-' ) {
					if ( (i == 8) || (i == 13) || (i == 18) || (i == 23) ) continue;
					nameIsUUID = false;
					break;
				} else {
					if ( (('0' <= ch) && (ch <= '9')) || (('a' <= ch) && (ch <= 'z')) ) continue;
					nameIsUUID = false;
					break;
				}
			}

		}
		
		if ( nameIsUUID ) {

			XMP_ExpandedXPath expPath;
			ExpandXPath ( kXMP_NS_XMP_MM, "InstanceID", &expPath );
			XMP_Node * idNode = FindNode ( tree, expPath, kXMP_CreateNodes, 0 );
			if ( idNode == 0 ) XMP_Throw ( "Failure creating xmpMM:InstanceID", kXMPErr_InternalFailure );

			idNode->options = 0;	// Clobber any existing xmpMM:InstanceID.
			idNode->value = tree->name;
			idNode->RemoveChildren();
			idNode->RemoveQualifiers();

			tree->name.erase();

		}

	}

}	// TouchUpDataModel


// -------------------------------------------------------------------------------------------------
// DetermineInputEncoding
// ----------------------
//
// We don't know the character encoding yet. This is complicated by the nasty case of single byte
// input. Try to accumulate at least 4 bytes of input. Throw an exception if the total input is
// less than 4 bytes. We need an outermost XML element, the shortest possible UTF-8 sequence is
// "<x/>", which still isn't even legit XMP. The result is the number of bytes absorbed.

static XMP_StringLen
DetermineInputEncoding ( XMLParserAdapter * xmlParser,
						 XMP_StringPtr	    buffer,
						 XMP_StringLen      xmpSize,
						 bool               lastClientCall )
{
	size_t copySize = 0;
	
	if ( (xmlParser->partialUniSize + xmpSize) < 4 ) {

		// Accumulate the input in the partialUniChar buffer.
		
		copySize = xmpSize;
		if ( lastClientCall ) XMP_Throw ( "Insufficient parsing input", kXMPErr_BadXML );
		for ( size_t i = 0; i < copySize; ++i ) {
			xmlParser->partialUniChar[xmlParser->partialUniSize+i] = buffer[i];
		}
		xmlParser->partialUniSize += copySize;

	} else {
	
		// Fill out the partialUniChar buffer and determine the encoding.
		
		copySize = 4 - xmlParser->partialUniSize;
		for ( size_t i = 0; i < copySize; ++i ) {
			xmlParser->partialUniChar[xmlParser->partialUniSize+i] = buffer[i];
		}
		xmlParser->partialUniSize += copySize;
		
		// The first character must be U+FEFF or ASCII, typically '<' for an outermost element,
		// initial processing instruction, or XML declaration. The second character can't be U+0000.
		// The possible input sequences are:
		//   Cases with U+FEFF
		//      EF BB BF -- - UTF-8
		//      FE FF -- -- - Big endian UTF-16
		//      00 00 FE FF - Big endian UTF 32
		//      FF FE 00 00 - Little endian UTF-32
		//      FF FE -- -- - Little endian UTF-16
		//   Cases with ASCII
		//      nn mm -- -- - UTF-8 -
		//      00 00 00 nn - Big endian UTF-32
		//      00 nn -- -- - Big endian UTF-16
		//      nn 00 00 00 - Little endian UTF-32
		//      nn 00 -- -- - Little endian UTF-16
		
		XMP_Uns8 * uniChar = (XMP_Uns8*)xmlParser->partialUniChar;	// ! Make sure comparisons are unsigned.
		
		if ( uniChar[0] == 0 ) {
		
			// These cases are:
			//   00 nn -- -- - Big endian UTF-16
			//   00 00 00 nn - Big endian UTF-32
			//   00 00 FE FF - Big endian UTF 32
			
			if ( uniChar[1] != 0 ) {
				xmlParser->charEncoding = kXMP_EncodeUTF16Big;			// 00 nn
			} else {
				if ( (uniChar[2] == 0) && (uniChar[3] != 0) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Big;		// 00 00 00 nn
				} else if ( (uniChar[2] == 0xFE) && (uniChar[3] == 0xFF) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Big;		// 00 00 FE FF
				}
			}
			
		} else {
		
			// These cases are:
			//   FE FF -- -- - Big endian UTF-16, FE isn't valid UTF-8
			//   FF FE 00 00 - Little endian UTF-32, FF isn't valid UTF-8
			//   FF FE -- -- - Little endian UTF-16
			//   nn mm -- -- - UTF-8, includes EF BB BF case
			//   nn 00 00 00 - Little endian UTF-32
			//   nn 00 -- -- - Little endian UTF-16
			
			if ( uniChar[0] == 0xFE ) {
				if ( uniChar[1] == 0xFF ) xmlParser->charEncoding = kXMP_EncodeUTF16Big;	// FE FF
			} else if ( uniChar[0] == 0xFF ) {
				if ( uniChar[1] == 0xFE ) {
					if ( (uniChar[2] == 0) && (uniChar[3] == 0) ) {
						xmlParser->charEncoding = kXMP_EncodeUTF32Little;	// FF FE 00 00
					} else {
						xmlParser->charEncoding = kXMP_EncodeUTF16Little;	// FF FE
					}
				}
			} else if ( uniChar[1] != 0 ) {
				xmlParser->charEncoding = kXMP_EncodeUTF8;					// nn mm
			} else {
				if ( (uniChar[2] == 0) && (uniChar[3] == 0) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Little;		// nn 00 00 00
				} else {
					xmlParser->charEncoding = kXMP_EncodeUTF16Little;		// nn 00
				}
			}

		}
		
		switch ( xmlParser->charEncoding ) {
			case kXMP_EncodeUTF16Big :
				xmlParser->UniConverter = (void*)UTF16BE_to_UTF8;
				break;
			case kXMP_EncodeUTF16Little :
				xmlParser->UniConverter = (void*)UTF16LE_to_UTF8;
				break;
			case kXMP_EncodeUTF32Big :
				xmlParser->UniConverter = (void*)UTF32BE_to_UTF8;
				break;
			case kXMP_EncodeUTF32Little :
				xmlParser->UniConverter = (void*)UTF32LE_to_UTF8;
				break;
			case kXMP_EncodeUTF8 :
				// Shift the buffered UTF-8 input so that it gets processed in the normal flow.
				for ( size_t i = 0; i < xmlParser->partialUniSize; ++i ) xmlParser->prevTail[i] = xmlParser->partialUniChar[i];
				xmlParser->prevTail[xmlParser->partialUniSize] = 0;
				xmlParser->partialUniSize = 0;
				break;
			default :
				XMP_Throw ( "Can't determine character encoding", kXMPErr_BadXMP );
		}
	
	}
	
	return copySize;
		
}	// DetermineInputEncoding


// -------------------------------------------------------------------------------------------------
// ProcessUTF8Portion
// ------------------
//
// The code to handle ASCII controls is vastly easier if the input is UTF-8. Higher level code
// converts the input buffers to UTF-8 if necessary.

static void
ProcessUTF8Portion ( XMLParserAdapter * xmlParser,
					 XMP_StringPtr	    buffer,
					 XMP_StringLen	    xmpSize,
					 XMP_OptionBits     options )
{
	// Old versions of the XMP toolkit wrote unescaped or singly escaped ASCII controls.
	// These get rejected by expat (or any other correct XML parser). Scan the input looking
	// for these and convert them to doubly escaped controls. We don't handle all possible
	// cases of escaping, just ones written by older versions of XMP: "&#xnn;" where "nn" is
	// one or two hex digits with uppercase letters. An incomplete escape at the end of a
	// buffer is saved as a string and finished from the start of the next buffer.
	
	// Ignore unescaped and singly escaped tab (0x9), LF (0xA), and CR (0xD), they are OK.

	size_t prevLen = strlen ( (char*) xmlParser->prevTail );
	XMP_Assert ( prevLen < 6 );
	size_t currLen = 0;	// The amount absorbed from the current buffer.

	while ( prevLen > 0 ) {
	
		// Have a partial escape left from the previous buffer. Gather 6 bytes and make a recursive
		// call to process it. Keep going until done or any leftover is fully in the current buffer.
		
		size_t moreLen = 6 - prevLen;
		if ( moreLen > xmpSize ) moreLen = xmpSize;
		for ( size_t i = 0; i < moreLen; ++i ) xmlParser->prevTail[prevLen+i] = buffer[i];
		xmlParser->prevTail[prevLen+moreLen] = 0;
		// printf ( "Leftover partial escape, had %d bytes, added %d bytes\n", prevLen, moreLen );

		currLen += moreLen;	// Accumulate the total amount taken from the current buffer;
		buffer  += moreLen;
		xmpSize -= moreLen;
		prevLen += moreLen;
		
		if ( prevLen < 6 ) {
			XMP_Assert ( xmpSize == 0 );
			break;	// ! Don't return, this might be the last buffer.
		}
		
		unsigned char  thisTail[kXMLParseTailSize];	// Capture the tail in a local buffer.
		XMP_Assert ( sizeof(thisTail) == sizeof(xmlParser->prevTail) );
		memcpy ( thisTail, xmlParser->prevTail, sizeof(thisTail) );	// AUDIT: Using dest sizeof for length is safe.
		xmlParser->prevTail[0] = 0;	// ! The recursive call must not think it has a tail.
		ProcessUTF8Portion ( xmlParser, (XMP_StringPtr)thisTail, prevLen, (options | kXMP_ParseMoreBuffers) );
		
		prevLen = strlen ( (char*) xmlParser->prevTail );	// Determine what was left from the recursive call.
		XMP_Assert ( prevLen < 6 );	// Need to make forward progress.
		if ( prevLen < currLen ) {	// Whatever is left is fully within this buffer.
			buffer  -= prevLen;
			xmpSize += prevLen;
			prevLen = 0;
			xmlParser->prevTail[0] = 0;
		}

	}
	
	XMP_Uns8 * spanStart = (XMP_Uns8*)buffer;
	XMP_Uns8 * bufferEnd = (XMP_Uns8*)buffer + xmpSize;
	XMP_Uns8 * spanEnd;
	XMP_Uns8 * refEnd;
	unsigned char controlValue;
	
	while ( spanStart < bufferEnd ) {
	
		// Find the next unescaped or singly escaped ASCII control.
		for ( spanEnd = spanStart; spanEnd < bufferEnd; ++spanEnd ) {

			if ( *((XMP_Uns8*)spanEnd) < 0x20 ) {

				refEnd = spanEnd + 1;
				if ( (*spanEnd != 0x9) && (*spanEnd != 0xA) && (*spanEnd != 0xD) ) break;

			} else if ( *spanEnd == '&' ) {

				refEnd = spanEnd + 1;
				controlValue = 0;

				if ( refEnd == bufferEnd ) break;	// A rather ugly sequence here, but hard to find better.
				if ( *refEnd == '#' ) {				// This is essentially a linearized state machine.
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( *refEnd == 'x' ) {
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( IsHexDigit ( *refEnd ) ) {
					controlValue = HexDigitValue ( *refEnd );
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( IsHexDigit ( *refEnd ) ) {	// ! The second hex digit is optional.
					controlValue = (controlValue << 4) + HexDigitValue ( *refEnd );
					if ( controlValue < 0x20 ) {
						++refEnd;
					} else {
						spanEnd = refEnd;
						continue;
					}
				}

				if ( refEnd == bufferEnd ) break;
				if ( (*refEnd == ';') && (controlValue != 0x9) && (controlValue != 0xA) && (controlValue != 0xD) ) {
					++refEnd;
					break;
				} else {
					spanEnd = refEnd;
					continue;
				}

			}

		}
		
		// Parse the span before the control, and the control if we found all of it. We get
		// here in one of three situations:
		//	1. The normal span extends to the end of the buffer.
		//	2. We have a normal span and a full control.
		//	3. We have a normal span and a partial control at the end of the buffer.
		
		// ! It is possible that refEnd == bufferEnd when we have a full control.

		if ( spanStart < spanEnd ) {
			// printf ( "Normal span: \"%.*s\"\n", (spanEnd - spanStart), spanStart ); fflush ( stdout );
			xmlParser->ParseBuffer ( spanStart, (spanEnd - spanStart), false );
			spanStart = spanEnd;
		}

		if ( spanEnd < bufferEnd ) {

			char escBuffer[32];
			XMP_Assert ( (*((XMP_Uns8*)spanEnd) < 0x20) || (*spanEnd == '&') );

			if ( *((XMP_Uns8*)spanEnd) < 0x20 ) {

				// printf ( "Raw control: 0x%X\n", *spanEnd ); fflush ( stdout );
				XMP_Assert ( refEnd == spanEnd+1 );
				snprintf ( escBuffer, sizeof(escBuffer), "&amp;#x%X;", *spanEnd );	// AUDIT: Using sizeof for snprintf length is safe.
				xmlParser->ParseBuffer ( escBuffer, strlen(escBuffer), false );

			} else {
			
				size_t refLen = refEnd - spanEnd;
				
				if ( *(refEnd-1) == ';' ) {
					++spanEnd;	// Skip the '&' in the singly escaped control.
					snprintf ( escBuffer, sizeof(escBuffer), "&amp;%.*s", refLen-1, spanEnd );	// AUDIT: Using sizeof for snprintf length is safe.
					// printf ( "Escaped control: %.*s -> %s\n", refLen, spanEnd-1, escBuffer ); fflush ( stdout );
					xmlParser->ParseBuffer ( escBuffer, strlen(escBuffer), false );
				} else {
					// printf ( "Partial escape: %.*s\n", refLen, spanEnd ); fflush ( stdout );
					XMP_Assert ( (refEnd == bufferEnd) && (xmlParser->prevTail[0] == 0) );
					XMP_Enforce ( refLen < sizeof(xmlParser->prevTail) );
					memcpy ( xmlParser->prevTail, spanEnd, refLen );	// AUDIT: Checked refLen above.
					xmlParser->prevTail[refLen] = 0;
				}

			}

			spanStart = refEnd;

		}
					
	}			
	
}	// ProcessUTF8Portion


// -------------------------------------------------------------------------------------------------
// ParseFromBuffer
// ---------------

void
XMPMeta::ParseFromBuffer ( XMP_StringPtr  buffer,
						   XMP_StringLen  xmpSize,
						   XMP_OptionBits options )
{
	if ( ! tree.children.empty() ) XMP_Throw ( "Can only parse into an empty object", kXMPErr_BadParse );
	if ( (buffer == 0) && (xmpSize != 0) ) XMP_Throw ( "Null parse buffer", kXMPErr_BadParam );
	if ( xmpSize == kXMP_UseNullTermination ) xmpSize = strlen ( buffer );
	
	const bool lastClientCall = ((options & kXMP_ParseMoreBuffers) == 0);	// *** Could use FlagIsSet & FlagIsClear macros.

	if ( xmlParser == 0 ) {
		if ( (xmpSize == 0) && lastClientCall ) return;	// Tolerate empty parse. Expat complains if there are no XML elements.
		xmlParser = new ExpatAdapter;
	}
	
	#if 0	// XMP_DebugBuild
		if ( xmlParser->parseLog != 0 ) {
			char message [200];
			snprintf ( message, sizeof(message), "<!-- ParseFromBuffer, length = %d, options = %X%s -->",
					   xmpSize, options, (lastClientCall ? " (last)" : "") );	// AUDIT: Using sizeof for snprintf length is safe.
			fwrite ( message, 1, strlen(message), xmlParser->parseLog );
			fflush ( xmlParser->parseLog );
		}
	#endif
	
	try {	// Cleanup the tree and xmlParser if anything fails.
	
		if ( xmlParser->charEncoding == XMP_OptionBits(-1) ) {
			size_t byteCount = DetermineInputEncoding ( xmlParser, buffer, xmpSize, lastClientCall );
			if ( xmlParser->charEncoding == XMP_OptionBits(-1) ) return;	// Still don't know the encoding.
			buffer  += byteCount;
			xmpSize -= byteCount;
		}
		
		if ( xmlParser->charEncoding == kXMP_EncodeUTF8 ) {

			// UTF-8 input buffers get directly processed.
			if ( xmpSize != 0 ) {	// The expat parser doesn't tolerate empty input.
				ProcessUTF8Portion ( xmlParser, buffer, xmpSize, options );
			}

		} else {
		
			// UTF-16 and UTF-32 input must be converted to UTF-8 then processed.
						
			enum { u8BufferSize = 4*1024 };
			XMP_Uns8 u8Buffer[u8BufferSize];
			size_t inCount, outCount;
			
			if ( xmlParser->partialUniSize > 0 ) {

				// We've got 1 to 3 bytes of a partial UTF-16 or UTF-32 character. Try to fill to 4
				// bytes, convert and process, resume below with the current input buffer. If we can
				// fill to 4 bytes then we're sure to get into the current buffer. UTF-32 input will
				// consume all 4 bytes. A UTF-16 surrogate pair will consume all 4 bytes. If the next
				// character is UTF-16 in the BMP we will consume at least 2 bytes. This happens only
				// if we enter with 1 byte, otherwise the BMP character would have been converted in
				// the prior call. If we can't fill to 4 bytes then we've consumed all of this buffer
				// and can leave any excess for the next call. There is also an edge case at the end
				// of UTF-6 input - come in with 1 byte, accumulate 1 more byte, have a BMP character.
				// So always try to convert and parse, even if we don't have 4 bytes.

				const XMP_StringPtr origBuffer = buffer;
				for ( ; (xmlParser->partialUniSize < 4) && (xmpSize > 0); ++xmlParser->partialUniSize, ++buffer, --xmpSize ) {
					xmlParser->partialUniChar[xmlParser->partialUniSize] = *buffer;
				}
				const size_t takenFromBuffer = buffer - origBuffer;
				
				if ( xmlParser->charEncoding & _XMP_UTF16_Bit ) {
					UTF16_to_UTF8_Proc UTF16_to_UTF8 = UTF16_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF16_to_UTF8 ( (UTF16Unit*)xmlParser->partialUniChar, xmlParser->partialUniSize/2, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 2;	// Convert it to a byte count.
				} else {
					XMP_Assert ( xmlParser->charEncoding & _XMP_UTF32_Bit );
					UTF32_to_UTF8_Proc UTF32_to_UTF8 = UTF32_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF32_to_UTF8 ( (UTF32Unit*)xmlParser->partialUniChar, xmlParser->partialUniSize/4, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 4;	// Convert it to a byte count.
				}
				
				const size_t partialLeft = xmlParser->partialUniSize - inCount;
				if ( partialLeft <= takenFromBuffer ) {
					buffer  -= partialLeft;
					xmpSize += partialLeft;
					xmlParser->partialUniSize = 0;
				} else {
					XMP_Assert ( xmpSize == 0 );
					for ( size_t i = 0; i < partialLeft; ++i ) xmlParser->partialUniChar[i] = xmlParser->partialUniChar[inCount+i];
					xmlParser->partialUniSize = partialLeft;
				}

			}
			
			// Convert and process the current buffer.
			
			bool doneLastPortion = false;
			while ( (xmpSize > 0) && (! doneLastPortion) ) {
			
				doneLastPortion = (xmpSize <= u8BufferSize/2);	// Worst case expansion is 2 bytes -> 3 bytes.
				
				if ( xmlParser->charEncoding & _XMP_UTF16_Bit ) {
					UTF16_to_UTF8_Proc UTF16_to_UTF8 = UTF16_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF16_to_UTF8 ( (UTF16Unit*)buffer, xmpSize/2, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 2;	// Convert it to a byte count.
				} else {
					XMP_Assert ( xmlParser->charEncoding & _XMP_UTF32_Bit );
					UTF32_to_UTF8_Proc UTF32_to_UTF8 = UTF32_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF32_to_UTF8 ( (UTF32Unit*)buffer, xmpSize/4, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 4;	// Convert it to a byte count.
				}
				
				buffer  += inCount;
				xmpSize -= inCount;
			
			}
			
			if ( xmpSize > 0 ) {
				XMP_Assert ( (xmpSize < 4) && (xmlParser->partialUniSize == 0) );
				for ( size_t i = 0; i < xmpSize; ++i ) xmlParser->partialUniChar[i] = buffer[i];
				xmlParser->partialUniSize = xmpSize;
				buffer += xmpSize;
				xmpSize = 0;
			}

		}
		
		if ( lastClientCall ) {
		
			if ( xmlParser->partialUniSize > 0 ) XMP_Throw ( "Input ends with a partial character", kXMPErr_BadXML );

			size_t remainder = strlen ( (char*)xmlParser->prevTail );
			// printf ( "Final remainder = %d\n", remainder ); fflush ( stdout );
			if ( remainder == 0 ) {
				xmlParser->prevTail[0] = ' ';	// The expat parser can't tolerate empty buffers.
				remainder = 1;
			}
			xmlParser->ParseBuffer ( (char*)xmlParser->prevTail, remainder, true );
			xmlParser->prevTail[0] = 0;
			
			#if XMP_DebugBuild && DumpXMLParseTree
				if ( xmlParser->parseLog == 0 ) xmlParser->parseLog = stdout;
				DumpXMLTree ( xmlParser->parseLog, xmlParser->tree, 0 );
			#endif

			const XML_Node * xmlRoot = FindRootNode ( this, *xmlParser, options );

			if ( xmlRoot != 0 ) {

				ProcessRDF ( &tree, *xmlRoot, options );
				NormalizeDCArrays ( &tree );
				if ( tree.options & kXMP_PropHasAliases ) MoveExplicitAliases ( &tree, options );
				TouchUpDataModel ( &tree );
				
				// Delete empty schema nodes. Do this last, other cleanup can make empty schema.
				size_t schemaNum = 0;
				while ( schemaNum < tree.children.size() ) {
					XMP_Node * currSchema = tree.children[schemaNum];
					if ( currSchema->children.size() > 0 ) {
						++schemaNum;
					} else {
						delete tree.children[schemaNum];	// ! Delete the schema node itself.
						tree.children.erase ( tree.children.begin() + schemaNum );
					}
				}
				
			}

			delete xmlParser;
			xmlParser = 0;

		}
		
	} catch ( ... ) {

		delete xmlParser;
		xmlParser = 0;
		prevTkVer = 0;
		tree.ClearNode();
		throw;

	}
	
}	// ParseFromBuffer

// =================================================================================================
